#!/usr/bin/python

# Author:  	Philip Cayting, Yale University
#			philip.cayting@yale.edu
# Function:  Along with SVdB.py, provides the backbone for BreakDB.  Event 
#  containers are defined here.  After querying the SVdB to get the data, Events
#  store the returned information.  Events also contain their display
#  information and are called by the display.py script.  Generally speaking, all
#  display code should be kept within the Event and all database access kept
#  within SVdB.

import SVdB

def exportSource(source_list):
	export = []
	for source in source_list:
		export.append(source['name'] + " PubMed(%s)" % source['pubmed_id'])
	return str(export)
	
## NOTE, FIX EVIDENCE HERE (Should have field for each evidence option)
EXPORT_FIELDS=[	("Event ID", "event_id"), ("Event Class", "event_class"),("Event Coordinates", "event_coordinates"),("Mechanism", "mechanism"),("Evidence", "evidence"),("Confidence Level", "confidence_level"),("Confidence Description", "confidence_description"),("Mechanism Notes", "mechanism_notes"),("Left Sequence", "left_sequence"),("Right Sequence", "right_sequence"),("Evidence Notes", "evidence_notes"),("Validation Notes", "validation_notes"), ("Species", "species"), ("Inserted Sequence", "inserted_sequence"), ("Related Events", "related_events"), ("Sources", "sources", exportSource)]

def combine(list_a, list_b):
	"Returns tuplized list combining lists a and b.  [(list_a[0], list_b[0]), ...]"
	assert len(list_a) == len(list_b)
	combined_list = []
	for i in range(0, len(list_a)):
		value = (list_a[i], list_b[i])
		combined_list.append(value)
	return combined_list


class SVEvent:

	def displayHTML(self, cell_color):
		display = "<table width=800 border=0>\n"
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Event ID")
		display += self.printTableData(cell_color, self.event_id)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printStartSection()
		display += self.printTableHeader("Event Class")
		display += self.printTableData(cell_color, self.event_class)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printEndRow()
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Event Coordinates")
		display += self.printTableData(cell_color, self.coordinatesString(self.event_coordinates, self.species, 0, self.event_id))
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printStartSection()
		display += self.printTableHeader("Confidence Level")
		display += self.printTableData(cell_color, self.confidence_level)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printEndRow()
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Confidence Description")
		display += self.printTableData(cell_color, self.confidence_description)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printStartSection()
		display += self.printTableHeader("Suggested Mechanism")
		display += self.printTableData(cell_color, self.mechanism)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printEndRow()
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Mechanism Notes")
		display += self.printTableData(cell_color, self.mechanism_notes)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printStartSection()
		display += self.printTableHeader("Left Sequence")
		display += self.printTableData(cell_color, self.left_sequence)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printEndRow()
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Right Sequence")
		display += self.printTableData(cell_color, self.right_sequence)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printStartSection()
		display += self.printTableHeader("Evidence")
		if self.evidence:
			display += self.printTableData(cell_color, self.evidence.display())
		else:
			display += self.printTableData(cell_color, "--")
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printEndRow()
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Evidence Notes")
		display += self.printTableData(cell_color, self.evidence_notes)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printStartSection()
		display += self.printTableHeader("Validation Notes")
		display += self.printTableData(cell_color, self.validation_notes)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printEndRow()
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Sources")
		display += self.printSources(cell_color, self.sources)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printStartSection()
		display += self.printTableHeader("Related Events")
		display += self.printRelatedVariants(cell_color, self.related_events)
		display += self.printBlankLine()
		display += self.printEndSection()
		display += self.printEndRow()
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Genome")
		display += self.printSpecies(cell_color, self.species)
		display += self.printBlankLine()
		display += self.printEndSection()
		if self.inserted_sequence:
			display += self.printStartSection()
			display += self.printTableHeader("Inserted Sequence")
			display += self.printTableData(cell_color, self.inserted_sequence)
			display += self.printBlankLine()
			display += self.printEndSection()
		display += self.printEndRow()
		
		display += self.printStartRow()
		display += self.printStartSection()
		display += self.printTableHeader("Supporting Reads")
		display += self.printReads(cell_color, self.read_pairs)
		display += self.printBlankLine()
		display += self.printEndSection()
		
		if hasattr(self, "archives"):
			display += self.printStartRow()
			display += self.printStartSection()
			display += self.printTableHeader("Archives")
			display += self.printArchives(cell_color, self.archives)
			display += self.printBlankLine()
			display += self.printEndSection()
			display += self.printEndRow()
			
		return display
		
	def printStartRow(self):
		return "<tr>"
	
	def printEndRow(self):
		return "</tr>"
		
	def printStartSection(self):
		return "<td valign=top style=\"word-break:break-all\"><table width=\"350\" border=0>"
		
	def printEndSection(self):
		return "</table></td>"
	
		
	def printTableHeader(self, title):
		return "<tr><td bgcolor=\"#CCCCCC\"><B>" + str(title) + "</B></td></tr>"
		
	def printTableData(self, cell_color, data):
		if not data: data = '--'
		return "<tr><td bgcolor=\"#%s\">%s</td></tr>" % (cell_color, data)
		
		
	def printBlankLine(self):
		return "<tr><td>&nbsp;</td></tr>"
		
	def printSpecies(self, cell_color, species):
		if species:
			display = "<tr><td bgcolor=\"#%s\">" % cell_color
			display += "%s (Build: %s)\n" % (species.name, species.build)
			display += "</td></tr>"
			return display
		else:
			return "<tr><td bgcolor=\"#%s\">--</td></tr>" % cell_color
			
	def printReads(self, cell_color, read_pairs):
		if read_pairs:
			display = "<tr><td bgcolor=\"#%s\"><ul>" % cell_color
			for rp in read_pairs:
				display += "<li><a href=\"read.py?id=%i\">%s</a>\n" % (rp.read_pair_id, rp.name)
			display += "</ul></td></tr>"
			return display
		else:
			return "<tr><td bgcolor=\"#%s\">--</td></tr>" % cell_color
						
	def coordinatesString(self, coords, species, count, name=""):
		if coords.chr:
			offset = int((coords.stop - coords.start)*.10)
			return "<ul style=\"list-style-type: none; list-style-position: inside; padding-left: %i;\"><li><a href=\"http://genome.ucsc.edu/cgi-bin/hgTracks?org=%s&db=%s&position=chr%s%%3A%i-%i&hgt.customText=http://sv.gersteinlab.org/ucsc/ucscFormat.py%%3Fchr%%3D%s%%26start%%3D%i%%26stop%%3D%i%%26name%%3D%s\">chr%s:%i-%i</a></ul>" % (count*8, species.name, species.ucsc_build(), coords.chr, coords.start-offset, coords.stop+offset, coords.chr, coords.start, coords.stop, name, coords.chr, coords.start, coords.stop)
		else:
			return "<ul style=\"list-style-type: none; list-style-position: inside; padding-left: %i;\"><li><a href=\"display.py?id=%s\">%s</a>:%i-%i\n%s</ul>" % (count*8, coords.context.event_id, coords.context.event_id, coords.start, coords.stop, self.coordinatesString(coords.context.event_coordinates, species, count+1, coords.context.event_id))
			
		
	def printArchives(self, cell_color, archives):
		if len(archives) > 1:
			display = "<tr><td bgcolor=\"#%s\">" % cell_color
			for archive in archives:
				display += "<a href=\"display.py?id=%s&archive=%i\">%s.%i</a> (%s)<BR>" % (archive.event_id, archive.archive_number, archive.event_id, archive.archive_number, archive.date_created.strftime('%m-%d-%Y'))
			display += "</td></tr>"
			return display
		else:
			return "<tr><td bgcolor=\"#%s\">--</td></tr>" % cell_color
			
	def printSources(self, cell_color, sources):
		if len(sources) == 0:
			return "<tr><td bgcolor=\"#%s\">--</td></tr>" % cell_color
		elif len(sources) == 1:
			return "<tr><td bgcolor=\"#%s\">%s, PubMed ID: %s</td></tr>" % (cell_color, sources[0]["name"], sources[0]["pubmed_id"])
		else:
			display = "<tr><td bgcolor=\"#%s\"><ul>" % cell_color
			for source in sources:
				display += "<li>%s, PubMedID: %s\n" % (source["name"], source["pubmed_id"])
			display += "</ul></td></tr>"
			return display
			
	def printRelatedVariants(self, cell_color, related_variants):
		if len(related_variants) == 0:
			return "<tr><td bgcolor=\"#%s\">--</td></tr>" % cell_color
		elif len(related_variants) == 1:
			return "<tr><td bgcolor=\"#%s\"><a href=\"display.py?id=%s\">%s</a></td></tr>" % (cell_color, related_variants[0], related_variants[0])
		else:
			display = "<tr><td bgcolor=\"#%s\"><ul>" % cell_color
			for variant in related_variants:
				display += "<li><a href=\"display.py?id=%s\">%s</a>\n" % (variant, variant)
			display += "</ul></td></tr>"
			return display
			
	def tabExport(self, fields):
		'Returns string of Variant data in tab delimited format intended for export'
		delimiter = '\t'
		export = []
		for field in fields:
			attribute = field[1]
			value = getattr(self, attribute)
			if value == None: value = ''
			if len(field) > 2:
				function = field[2]
				value = function(value)  # Apply custom (optional) function
			elif not isinstance(value, str):
				value = value.__str__()  # Default obj->string function
			export.append(value)
		return delimiter.join(export)
						

class ExistingSVEvent(SVEvent):
	def __init__(self, id, conn):
		self.event_id = id
		
		# Ask SVdB for Event Information
		db = SVdB.DBActiveSVEvent(id, conn)
		
		self.archive_id = db.archive_number
		self.event_class = db.event_class
		self.event_coordinates = db.event_coordinates
		self.archives = db.archives
		self.mechanism = db.mechanism
		self.evidence = db.evidence
		self.confidence_level = db.confidence_level
		self.confidence_description = db.confidence_description
		self.mechanism_notes = db.mechanism_notes
		self.left_sequence = db.left_sequence
		self.right_sequence = db.right_sequence
		self.evidence_notes = db.evidence_notes
		self.validation_notes = db.validation_notes
		self.date_created = db.date_created
		self.sources = db.sources
		self.related_events = db.related_events
		self.species = db.species
		self.inserted_sequence = db.inserted_sequence
		self.read_pairs = SVdB.getReadPairs(id, conn)
		
	def displayEditHTML(self, conn, cell_color):
		display = "<table width=800 border=0>\n"
		
		display += self.printTableHeader('Event ID')
		display += self.printTableData(cell_color, self.event_id)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Event Class')
		display += self.printTableData(cell_color, self.event_class)
		display += self.printEditList(cell_color, 'event_class', self.event_class, SVdB.getValidEventClasses(conn))
		display += self.printBlankLine()
		
		display += self.printTableHeader('Event Coordinates')
		display += self.printTableData(cell_color, self.coordinatesString(self.event_coordinates, self.species, 0, self.event_id))
		display += self.printEditCoordinates(cell_color, self.event_coordinates)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Confidence Level')
		display += self.printTableData(cell_color, self.confidence_level)
		display += self.printEditField(cell_color, 'confidence_level', self.confidence_level)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Confidence Description')
		display += self.printTableData(cell_color, self.confidence_description)
		display += self.printEditTextBox(cell_color, 'confidence_description', self.confidence_description)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Suggested Mechanism')
		display += self.printTableData(cell_color, self.mechanism)
		display += self.printEditList(cell_color, 'mechanism', self.mechanism, SVdB.getValidMechanisms(conn))
		display += self.printBlankLine()
		
		display += self.printTableHeader('Mechanism Notes')
		display += self.printTableData(cell_color, self.mechanism_notes)
		display += self.printEditTextBox(cell_color, 'mechanism_notes', self.mechanism_notes)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Left Sequence')
		display += self.printTableData(cell_color, self.left_sequence)
		display += self.printEditTextBox(cell_color, 'left_sequence', self.left_sequence)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Right Sequence')
		display += self.printTableData(cell_color, self.right_sequence)
		display += self.printEditTextBox(cell_color, 'right_sequence', self.right_sequence)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Evidence')
		if self.evidence:
			display += self.printTableData(cell_color, self.evidence.display())
		else:
			display += self.printTableData(cell_color, "None")
		display += self.printEvidenceEditHTML(cell_color, self.evidence)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Evidence Notes')
		display += self.printTableData(cell_color, self.evidence_notes)
		display += self.printEditTextBox(cell_color, 'evidence_notes', self.evidence_notes)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Validation Notes')
		display += self.printTableData(cell_color, self.validation_notes)
		display += self.printEditTextBox(cell_color, 'validation_notes', self.validation_notes)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Related Events')
		related_ids = self.related_events
		display += self.printTableData(cell_color, related_ids)
		display += self.printEditTextBox(cell_color, 'related_events', related_ids)
		display += self.printTableData(cell_color, "(Separate IDs by commas)")
		display += self.printBlankLine()
		
		display += self.printTableHeader('Sources')
		sources = map(lambda x: (x['name'], x['pubmed_id']), self.sources)
		display += self.printTableData(cell_color, sources)
		display += self.printEditSources(cell_color, 'sources', sources, 5)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Genome')
		species = SVdB.allSpecies(conn)
		option_list = map(lambda x: x.species_id, species)
		name_list = map(lambda x: "%s (Build: %s)" % (x.name, x.build), species)
		if self.species:
			display += self.printTableData(cell_color, self.species.display())
			display += self.printEditList(cell_color, 'species', self.species.species_id, option_list, name_list)
		else:
			display += self.printTableData(cell_color, '--')
			display += self.printEditList(cell_color, 'species', 0, option_list, name_list)
		display += self.printBlankLine()
		
		display += self.printTableHeader('Inserted Sequence')
		display += self.printTableData(cell_color, self.inserted_sequence)
		display += self.printEditTextBox(cell_color, 'inserted_sequence', self.inserted_sequence)
		display += self.printBlankLine()
		
		
		display += "</table>"
		
		return display
		
	def printEditSources(self, cell_color, field_name, data_list, num_items):
		display = "<tr><td bgcolor=\"#%s\">" % cell_color
		for i in range(0, num_items):
			source = ''
			pubmed_id = ''
			if len(data_list) > i and data_list[i]:
				source = data_list[i][0]
				pubmed_id = data_list[i][1]
			display += "Source: <input type=\"text\" name=\"%s_source%i\" value=\"%s\"> " % (field_name, i, source)
			display += "&nbsp;PubMed ID: <input type=\"text\" name=\"%s_pubmed%i\" value=\"%s\"><BR>" % (field_name, i, pubmed_id)
		display += "</td></tr>"
		return display
		
	def printEditList(self, cell_color, field_name, data, option_list, name_list=[]):
		if name_list == []:  name_list = option_list
		if not data: data = ''
		display = "<tr><td bgcolor=\"#%s\"><SELECT name=\"%s\">\n" % (cell_color, field_name)
		for option, name in combine(option_list, name_list):
			if option == data:
				display += "<OPTION SELECTED value=\"" + str(option) + "\">" + str(name) + "</OPTION>"
			else:
				display += "<OPTION value=\"" + str(option) + "\">" + str(name) + "</OPTION>"
		display += "</td></tr>"
		
		return display
		
	def printEditField(self, cell_color, field_name, data):
		if not data: data = ''
		return "<tr><td bgcolor=\"#%s\"><input type=\"text\" name=\"%s\" value=\"%s\"></td></tr>" % (cell_color, field_name, str(data))
		
	def printEditTextBox(self, cell_color, field_name, data):
		if not data: data = ''
		if isinstance(data, list):
			display = "<tr><td bgcolor=\"#%s\"><textarea rows=\"8\" cols=\"60\" name=\"%s\">" % (cell_color, field_name)
			for i, item in enumerate(data):
				if i == 0: display += str(item)
				else:  display += ", " + str(item) 
			display += "</textarea></td></tr>"
			return display
		else:
			return "<tr><td bgcolor=\"#%s\"><textarea rows=\"8\" cols=\"60\" name=\"%s\">%s</textarea></td></tr>" % (cell_color, field_name, str(data))
		
	def printEvidenceEditHTML(self, cell_color, evidence):
		primate_sequence = ''
		homology_length = ''
		homology_percentage = ''
		if evidence:
			primate_sequence = evidence.primate_sequence
			homology_length = evidence.homology_length
			homology_percentaget = evidence.homology_percentage
	
		html = "<tr><td bgcolor=\"#%s\">Sequence Found in Primates:  <input type=\"checkbox\" name=\"primate_sequence\" " % (cell_color)
		if primate_sequence:
			html += "CHECKED "
		html += "><BR>"
		html += "Homology Length:  <input type=\"text\" name=\"homology_length\" value=\"%s\"><BR>" % (homology_length)
		html += "Homology Percentage:  <input type=\"text\" name=\"homology_percentage\" value=\"%s\"><BR>" % (homology_percentage)
		html += "</td></tr>"
		return html
		
	def printEditCoordinates(self, cell_color, coordinates):
		display = "<tr><td bgcolor=\"#%s\">" % (cell_color)
		if coordinates.context and isinstance(coordinates.context, SVEvent):
			display += "Parent Event: <input type=\"text\" name=\"context_id\" value=\"%s\"><BR>" % (coordinates.context.event_id)
		else:
			display += "Parent Event: <input type=\"text\" name=\"context_id\"><BR>"
		display += "Start Coordinate: <input type=\"text\" name=\"start_coordinate\" value=\"%i\">&nbsp; Stop Coordinate: <input type=\"text\" name=\"stop_coordinate\" value=\"%i\"><BR>" % (coordinates.start, coordinates.stop)
		display += "Chromosome: <input type=\"text\" name=\"chromosome\" value=\"%s\">" % coordinates.chr
		
		return display
		
		
class NewSVEvent(SVEvent):
	def __init__(self, event_id=0, event_class=None, event_coordinates=None, mechanism=None, evidence=None, confidence_level=None, confidence_description=None, mechanism_notes=None, left_sequence=None, right_sequence=None, evidence_notes=None, validation_notes=None, sources=[], related_events=[], species=None, inserted_sequence=None, read_pairs=[]):
		self.event_id = event_id
		self.event_class = event_class
		self.event_coordinates = event_coordinates
		self.mechanism = mechanism
		self.evidence = evidence
		self.confidence_level = confidence_level
		self.confidence_description = confidence_description
		self.mechanism_notes = mechanism_notes
		self.left_sequence = left_sequence
		self.right_sequence = right_sequence
		self.evidence_notes = evidence_notes
		self.validation_notes = validation_notes
		self.sources = sources
		self.related_events = related_events
		self.species = species
		self.inserted_sequence = inserted_sequence
		self.read_pairs = read_pairs
		
			
class ArchiveSVEvent(SVEvent):
	def __init__(self, event_id, archive_number, conn):
		self.archive_number = archive_number
		
		# Ask SVdB for Archive Variant Information
		db = SVdB.DBArchiveSVEvent(event_id, archive_number, conn)
		
		self.event_id = db.event_id
		self.event_class = db.event_class
		self.event_coordinates = db.event_coordinates
		self.mechanism = db.mechanism
		self.evidence = db.evidence
		self.confidence_level = db.confidence_level
		self.confidence_description = db.confidence_description
		self.mechanism_notes = db.mechanism_notes
		self.left_sequence = db.left_sequence
		self.right_sequence = db.right_sequence
		self.evidence_notes = db.evidence_notes
		self.validation_notes = db.validation_notes
		self.date_created = db.date_created
		self.related_events = db.related_events
		self.sources = db.sources
		self.species = db.species
		self.inserted_sequence = db.inserted_sequence
		self.read_pairs = SVdB.getReadPairs(self.event_id, conn)
		

		


		
		
								   
								   
		