#ifndef __DNAFRAGMENT_HH__
#define __DNAFRAGMENT_HH__

#include "Genome.hh"
#include "GenomeShifter.hh"

TString SEQ_LINKER("GTTGGAACCGAAAGGGTTTGAATTCAAACCCTTTCGGTTCCAAC");  

class DNAFragment : public Region
{
  // Constructors and detructors
private:
  DNAFragment(int start,int end,TString seq) : Region(start,end),seq_(seq)
  {}

public:
  DNAFragment(TString seq) : Region(-seq.Length(),-1), seq_(seq)
  {}

public:
  ~DNAFragment ()
  {}

private:

  // Fragment generator
public:
  static DNAFragment *generateFragment(Genome *genome,
				       double logmean,double logsd)
  {
    int g_start = genome->getStart();
    int g_end   = genome->getEnd();

    int start = -1, end = -1, g_len = g_end - g_start + 1;
    while (!(start >= g_start && start <= g_end &&
	     end   >= g_start && end   <= g_end)) {
      double rand = gRandom->Rndm();
      int    mean = g_start + int(rand*g_len);
      int    len  = int(TMath::Exp(gRandom->Gaus(logmean,logsd)));
      int    half = len>>1;
      start   = mean - half;
      end     = start + len - 1;
    }

    DNAFragment *ret = new DNAFragment(start,end,
				       genome->getSequence(start,end));
    return ret;
  }

  // Sequence
private:
  TString seq_;
public:
  TString getSequence() { return seq_; }
  void    setSequence(TString seq) { seq_ = seq; }
  TString getLinker() { return SEQ_LINKER; }

  // First end
private:
  int pe_start1_, pe_end1_, pe_len1_;
  TString pe_seq1_;
public:
  int getEndStart1()    { return pe_start1_; }
  int getEndEnd1()      { return pe_end1_; }
  int getEndLength1()   { return pe_len1_; }
  TString getEndSeq1()  { return pe_seq1_; }

  // Second end
private:
  int pe_start2_, pe_end2_, pe_len2_;
  TString pe_seq2_;
public:
  int getEndStart2()   { return pe_start2_; }
  int getEndEnd2()     { return pe_end2_; }
  int getEndLength2()  { return pe_len2_; }
  TString getEndSeq2() { return pe_seq2_; }

  // Sequencing read for 454
private:
  int read_len_;
  TString read_seq_;
  int read_len2_;
  TString read_seq2_;
public:
  int get454ReadLength()   { return read_len_; }
  TString get454ReadSeq()  { return read_seq_; }
  int getSolexaReadLength1()  { return read_len_; }
  TString getSolexaReadSeq1() { return read_seq_; }
  int getSolexaReadLength2()  { return read_len2_; }
  TString getSolexaReadSeq2() { return read_seq2_; }


  // Generation of reads and ends for Solexa technology
public:
  bool makeSolexaReads(LengthDistribution *reads_distr)
  {
    read_len_ = read_len2_ = 0;
    read_seq_ = read_seq2_ = "";

    if (!seq_ || seq_ == "") {
      cerr<<"ERROR: No sequence is set."<<endl;
      return false;
    }

    if (seq_.Length() != getLength()) {
      cerr<<"ERROR: Sequence length ("<<seq_.Length()<<") does not match the "
	  <<"fragmet length ("<<getLength()<<")."<<endl;
      cerr<<"No ends generated."<<endl;
      return false;
    }
    
    pe_len1_  = reads_distr->generateLength();
    pe_len2_ = reads_distr->generateLength();
    if (pe_len1_ + pe_len2_ > getLength()) {
      cerr<<"Generated read lengths ("<<pe_len1_<<", "<<pe_len2_
	  <<") are longer than fragment."<<endl;
      return false;
    }

    pe_start1_ = 0;
    pe_end1_   = pe_len1_ - 1;
    pe_seq1_   = seq_(pe_start1_,pe_len1_);

    pe_start2_ = getLength() - pe_len2_;
    pe_end2_   = getLength() - 1;
    pe_seq2_   = seq_(pe_start2_,pe_len2_);

    read_len_ = pe_len1_;
    read_seq_ = pe_seq1_;

    read_len2_ = pe_len2_;
    for (int i = pe_len2_ - 1;i >= 0;i--) {
      char c = pe_seq2_[i];
      if (c == 'A')      read_seq2_ += 'T';
      else if (c == 'C') read_seq2_ += 'G';
      else if (c == 'T') read_seq2_ += 'A';
      else if (c == 'G') read_seq2_ += 'C';
      else               read_seq2_ += 'X';
    }

    return true;
  }

  // Generation of read and ends for 454 technology
public:
  bool make454Read(LengthDistribution *reads_distr)
  {
    read_len_ = read_len2_ = 0;
    read_seq_ = read_seq2_ = "";

    if (!seq_ || seq_ == "") {
      cerr<<"ERROR: No sequence is set."<<endl;
      return false;
    }

    if (seq_.Length() != getLength()) {
      cerr<<"ERROR: Sequence length ("<<seq_.Length()<<") does not match the "
	  <<"fragment length ("<<getLength()<<")."<<endl;
      cerr<<"No ends generated."<<endl;
      return false;
    }

    read_len_ = reads_distr->generateLength();
    if (read_len_ < SEQ_LINKER.Length() + 1) {
      cerr<<"Generated read length ("<<read_len_
	  <<") is shorter than linker."<<endl;
      return false;
    }

    int len = SEQ_LINKER.Length(), half = len>>1;
    int start = -1, end = -1, min = -1;
//     while (!(start >= 0 && start < read_len_ &&
// 	     end   >= 0 && end   < read_len_ &&
// 	     min   >= 22)) {
//       double rand = gRandom->Rndm();
//       int    mean = int(rand*read_len_);
//       start = mean - half;
//       end   = start + len - 1;
//       min   = read_len_ - end - 1;
//       if (start < min) min = start;
//       if (double(min - 22)/double(37 - 22) < gRandom->Rndm()) min = 0;
//     }
    while (!(start >= 0 && start < read_len_ &&
	     end   >= 0 && end   < read_len_)) {
      double rand = gRandom->Rndm();
      int    mean = int(rand*read_len_);
      start = mean - half;
      end   = start + len - 1;
    }

    pe_start1_ = 0;
    pe_end1_   = start - 1;
    pe_len1_   = pe_end1_ - pe_start1_ + 1;
    pe_seq1_   = seq_(pe_start1_,pe_end1_ - pe_start1_ + 1);

    pe_len2_   = read_len_ - pe_len1_ - len;
    pe_start2_ = getLength() - pe_len2_;
    pe_end2_   = getLength() - 1;

    pe_seq2_   = seq_(pe_start2_,pe_end2_ - pe_start2_ + 1);

    read_seq_ = pe_seq2_ + SEQ_LINKER + pe_seq1_;

    return true;
  }

public:
  TString make454Header(int readIndex,GenomeShifter *shifter)
  {
    int startA = getStart() + getEndStart1();
    int endA   = getStart() + getEndEnd1();
    int startB = getStart() + getEndStart2();
    int endB   = getStart() + getEndEnd2();
    if (shifter) shifter->shift(startA,endA,startB,endB);
    
    TString read = get454ReadSeq();
    
    TString ret = ">";
    ret += readIndex;
    ret += "_"; ret += startB;
    ret += "_"; ret += endB;
    ret += "_"; ret += startA;
    ret += "_"; ret += endA;
    ret += " length=";  ret += read.Length();
    ret += " uaccno=E"; ret += readIndex;
    
    return ret;
  }

  TString makeSolexaHeader1(int readIndex,GenomeShifter *shifter)
  {
    int start = getStart() + getEndStart1();
    int end   = getStart() + getEndEnd1();
    TString read = getSolexaReadSeq1();
    TString beg = ">";
    beg += readIndex;
    beg += "_1";
    return makeSolexaHeader(beg,readIndex,read,start,end,shifter);
  }

  TString makeSolexaHeader2(int readIndex,GenomeShifter *shifter)
  {
    int start = getStart() + getEndStart2();
    int end   = getStart() + getEndEnd2();
    TString read = getSolexaReadSeq2();
    TString beg = ">";
    beg += readIndex;
    beg += "_2";
    return makeSolexaHeader(beg,readIndex,read,start,end,shifter);
  }

private:
  TString makeSolexaHeader(TString beg,int readIndex,TString read,
			   int start,int end,
			   GenomeShifter *shifter)
  {
    TString ret = beg;
    if (shifter) shifter->shift(start,end);
    ret += "_"; ret += start;
    ret += "_"; ret += end;
    ret += " length=";  ret += read.Length();
    ret += " uaccno=E"; ret += readIndex;
    
    return ret;
    
  }

};

#endif
