#include "Sequenator.hh"

Sequenator::Sequenator()
{
  bg_mean_      = 0.20;
  bg_sigma_     = 0.1;
  sig_sigma_[1] = 0.17;

  logmean_  = log(bg_mean_*bg_mean_/sqrt(bg_mean_*bg_mean_ +
					 bg_sigma_*bg_sigma_));
  logsigma_ = sqrt(log(bg_sigma_*bg_sigma_/(bg_mean_*bg_mean_) + 1));

  setSignalSigmaSquare();
  //setSignalSigmaLinear();
  calcCutOffs();
  reset();
}

Sequenator::Sequenator(double bg_mean,
		       double bg_sigma,
		       double sig_sigma) : bg_mean_(bg_mean),
					   bg_sigma_(bg_sigma)
{
  sig_sigma_[1] = sig_sigma;

  logmean_  =
    log(bg_mean_*bg_mean_/sqrt(bg_mean_*bg_mean_ +
			       bg_sigma_*bg_sigma_));
  logsigma_ =
    sqrt(log(bg_sigma_*bg_sigma_/(bg_mean_*bg_mean_) + 1));

  setSignalSigmaSquare();
  //setSignalSigmaLinear();
  calcCutOffs();
  reset();
}

void Sequenator::calcCutOffs()
{
  for (int i = 0;i < N_POLYMERS;i++) {
    cutoff[i][0] = i;
    cutoff[i][1] = i + 1;
  }
  for (int i = 1;i < N_POLYMERS;i++) {
    double xu = i, xl = (i - 1);
    if (i == 1) xl = bg_mean_;
    while (xu - xl > 0.0001) {
      double x = (xu + xl)/2.;
      double vu = getSignalDistributionValue(i,x), vl = 0;
      if (i == 1) vl = getBGDistributionValue(x);
      else        vl = getSignalDistributionValue(i - 1,x);
      if (vl > vu)      xl = x;
      else if (vl < vu) xu = x;
      else xl = xu = x;
    }
    double x = (xu + xl)/2.;
    cutoff[i - 1][1] = cutoff[i][0] = x;
  }
    
//   for (int i = 0;i < N_POLYMERS;i++)
//     cout<<"Cut offs for "<<i<<" are: "
// 	<<cutoff[i][0]<<"-"<<cutoff[i][1]<<endl;
}

void Sequenator::setSignalSigmaLinear()
{
  for (int i = 2;i < N_POLYMERS;i++)
    sig_sigma_[i] = sig_sigma_[i - 1] + sig_sigma_[1];
}

void Sequenator::setSignalSigmaSquare()
{
  for (int i = 2;i < N_POLYMERS;i++)
    sig_sigma_[i] = sqrt(i)*sig_sigma_[1];
}

double Sequenator::getBGDistributionValue(double x)
{
  static const double twopisqrt = 1./sqrt(2.*M_PI);
  if (x < bg_mean_) return 0;
  double logx = log(x);
  double logsigma_over = 1./logsigma_;
  double val = (logx - logmean_)*logsigma_over;
  return twopisqrt*logsigma_over*exp(-0.5*val*val)/x;
}

double Sequenator::getSignalDistributionValue(int n_bp,double x)
{
  static const double twopisqrt = 1./sqrt(2.*M_PI);
  double sigma_over = 1./signalSigma(n_bp);
  double val = (x - n_bp)*sigma_over;
  return twopisqrt*sigma_over*exp(-0.5*val*val);
}


int Sequenator::getNCalled(double val)
{
  if (val < cutoff[1][0]) return 0;

  for (int i = 1;i < N_POLYMERS;i++)
    if (val < cutoff[i][1]) return i;

  return 0;
}


void Sequenator::reset() 
{
  for (int i = 0;i < N_POLYMERS;i++)
    poly_distr_[i] = ins_distr_[i] = del_distr_[i] = sub_distr_[i] = 0;
}


void Sequenator::addSequenced(int n)
{
  if (n >= 0 && n < N_POLYMERS)
    poly_distr_[n] += n;
}

void Sequenator::addInserted(int ind,int n)
{
  if (ind >= 0 && ind < N_POLYMERS)
    ins_distr_[ind] += n;
}

void Sequenator::addDeleted(int ind,int n)
{
  if (ind >= 0 && ind < N_POLYMERS)
    del_distr_[ind] += n;
}

void Sequenator::addSubstituted(int ind,int n)
{
  if (ind >= 0 && ind < N_POLYMERS)
    sub_distr_[ind] += n;
}

TString Sequenator::sequence454(TString seq)
{
  char flows[] = {'T','A','C','G','T','A','C','G'};
  int flow_ind = 0;
  TString ret = "";

  char curr_c = seq(0);
  int  curr_n = 1;
  seq += '-';
  for (int i = 1;i < seq.Length();i++) {

    // Constructing homopolymer
    char c = seq(i);
    if (c == curr_c) {
      curr_n++;
      if (i < seq.Length()) continue;
    }

    // Check for characters that are not ACTG
    addSequenced(curr_n);
    int n_read = 0;
    if (curr_c != flows[0] && curr_c != flows[1] &&
	curr_c != flows[2] && curr_c != flows[3]) {
      n_read = curr_n;
    } else {
      char flow_c = flows[flow_ind++];
      while (curr_c != flow_c) { // Flow is different -- generate background
	double val = generate454BGValue();
	n_read = getNCalled(val);
	addInserted(0,n_read);
	while (n_read > 0) {
	  ret += flow_c;
	  n_read--;
	}
	flow_c = flows[flow_ind++];
      }

      if (flow_ind >= 4) flow_ind -= 4;

    // Same flow, generate homopolymer size
      double val = generate454SignalValue(curr_n);
      n_read = getNCalled(val);
      int delta = curr_n - n_read;
      if (delta > 0) addDeleted(curr_n,delta);
      else if (delta < 0) addInserted(curr_n,-delta);
    }

    // Add sequenced homopolymer to the sequence
    while (n_read > 0) {
      ret += curr_c;
      n_read--;
    }

    curr_c = c;
    curr_n = 1;
  }
  return ret;
}

TString Sequenator::sequenceSolexa(TString seq)
{
  char nucs[] = {'T','A','C','G'};

  TString ret = "";

  int n_err = 0;
  for (int i = 0;i < seq.Length();i++) {

    char c = seq(i);
    addSequenced(1);

    // Input check
    if (c != nucs[0] && c != nucs[1] &&
	c != nucs[2] && c != nucs[3]) {
      ret += c;
      continue;
    }

    int i2 = (i + 1)*(i + 1);

    // Error rate estimation from Nucleic Acids Res. 2008 September; 36(16): e105.
    double error_rate = 3e-3 + 3.3e-8*i2*i2;
    if (error_rate > 0.75) error_rate = 0.75;

    // Sequencing
    if (gRandom->Uniform() >= error_rate) { // No error
      ret += c;
    } else { // Generate random nucleotide
      int index = int(gRandom->Uniform(0,2.9999999));
      if (c == nucs[0]      && 0 <= index) index++;
      else if (c == nucs[1] && 1 <= index) index++;
      else if (c == nucs[2] && 2 <= index) index++;
      ret += nucs[index];
      n_err++;
    }
  }

  addSubstituted(1,n_err);

  return ret;
}
