//--- C/C++ includes ---
#include <iostream>
#include <fstream>
#include <math.h>
using namespace std;

//--- ROOT includes ---
#include <TString.h>
#include <TROOT.h>
#include <TH1D.h>
#include <TFile.h>
#include <TRandom3.h>
#include <TMath.h>

TROOT root("Rint","The ROOT Interactive Interface");

//--- Application includes ---
#include "LengthDistribution.hh"
#include "Region.hh"
#include "DNAFragment.hh"
#include "SVEvent.hh"
#include "Genome.hh"
#include "Sequenator.hh"
#include "GenomeShifter.hh"

TString constructChimericSequence(DNAFragment *frg1,DNAFragment *frg2);
TString complement(TString seq);

int main(int argc,char *argv[])
{
  double logmean  = 7.8;
  double logsd    = 0.29;
  int    coverage = 7;
  bool   diploid  = false;
  bool   use454   = false, useSolexa = false, chimeric = false;

  TString usage = "usage:\n";
  usage += "\tsim -rd rd_file.txt or value [-svd svd_file.txt]\n";
  usage += "\t\t[-diploid] [-logmean val:default=7.8]\n";
  usage += "\t\t[-logsd val:default=0.29] [-coverage val:default=7] \n";
  usage += "\t\t[-454 (default) | -solexa] genome.fa\n";
  if (argc <= 1) {
    cerr<<usage<<endl;
    return 1;
  }

  TString svd_file = "", rd_file = "", genome_file = "";

  for (int i = 1;i < argc;i++) {
    TString val = argv[i];
    if      (val == "-rd"  && i + 1 < argc && argv[i + 1][0] != '-')
      rd_file = argv[++i];
    else if (val == "-svd" && i + 1 < argc && argv[i + 1][0] != '-')
      svd_file = argv[++i];
    else if (val == "-diploid") diploid = true;
    else if (val == "-logmean" && i + 1 < argc && argv[i + 1][0] != '-')
      logmean = strtof(argv[++i],NULL);
    else if (val == "-logsd" && i + 1 < argc && argv[i + 1][0] != '-')
      logsd = strtof(argv[++i],NULL);
    else if (val == "-coverage" && i + 1 < argc && argv[i + 1][0] != '-')
      coverage = strtol(argv[++i],NULL,10);
    else if (val == "-454")    use454 = true;
    else if (val == "-solexa") useSolexa = true;
    else if (val(0) == '-') {
      cerr<<"Unknown option '"<<val<<"'."<<endl;
      cerr<<usage<<endl;
      return 1;
    } else if (genome_file.Length() != 0) {
      cerr<<"Can use only one input file."<<endl;
      cerr<<usage<<endl;
      return 1;
    } else {
      genome_file = argv[i];
    }
  }

  if (useSolexa && use454) {
    cerr<<"Only one sequencing technology can be used."<<endl;
    cerr<<usage<<endl;
    return 1;
  }

  if (!useSolexa && !use454) use454 = true; // Default
  
  if (rd_file.Length() == 0 || genome_file.Length() == 0) {
    cerr<<"Bad input string."<<endl;
    cerr<<usage<<endl;
    return 1;
  }

  cout<<"Intend to generate "<<coverage<<"x coverage "
      <<"with DNA fragments of length "<<int(exp(logmean))<<"."<<endl<<endl;
  cout<<"Parameters of lognormal distribution for DNA fragments are:"<<endl;
  cout<<"logmean = "<<logmean<<endl;
  cout<<"logsd   = "<<logsd<<endl;

  int arr[10000*3];
  int n_sv_lines = 0;
  if (svd_file.Length() != 0)
    n_sv_lines = SVEvent::readEvents(svd_file,arr);

  int read_length = strtol(rd_file,NULL,10);
  LengthDistribution *reads_distr = new LengthDistribution(read_length);
  if (read_length <= 0) {
    delete reads_distr;
    reads_distr = new LengthDistribution(rd_file);
  }
  
  Genome ref_genome(genome_file);
  if (!ref_genome.isFine()) {
    cerr<<ref_genome.getErrorMessage()<<endl;
    return 1;
  }

  int start = ref_genome.getStart();
  int end   = ref_genome.getEnd();
  int len   = ref_genome.getLength();
  cout<<"Reference genome is chr"<<ref_genome.getChromosome()<<":"
      <<start<<"-"<<end<<endl;
  cout<<"It is of "<<ref_genome.getLength()<<" nucleotides in length."<<endl;
  cout<<"Chunked in "<<ref_genome.getNChunks()<<" pieces."<<endl<<endl;


  // Setting up random generator
  gRandom = new TRandom3();
  cout<<"Using "<<gRandom->ClassName()<<" random generator."<<endl<<endl;

  // Creating histograms
  int rUp   = reads_distr->getUpRange();
  int rLow  = reads_distr->getLowRange();
  TH1 *h_sv_start  = new TH1D("h_sv_start", "SV start", 10000,start,end);
  TH1 *h_sv_end    = new TH1D("h_sv_end  ", "SV end  ", 10000,start,end);
  TH1 *h_sv_len    = new TH1D("h_sv_len  ", "SV len  ", 10000,0.5,10000.5);
  TH1 *h_end_len   = new TH1D("h_end_len  ","End len  ",rUp,0.5,rUp + 0.5);
  TH1 *h_read_len  = new TH1D("h_read_len", "Read len",
			      rUp - rLow + 1,rLow - 0.5,rUp + 0.5);
  TH1 *h_sig_null  = new TH1D("h_sig_null", "0 nuc",     500,0,10);
  TH1 *h_sig_1     = new TH1D("h_sig_1",    "1 nuc",     500,0,10);
  TH1 *h_sig_2     = new TH1D("h_sig_2",    "2 nuc",     500,0,10);
  TH1 *h_sig_3     = new TH1D("h_sig_3",    "3 nuc",     500,0,10);
  TH1 *h_sig_4     = new TH1D("h_sig_4",    "4 nuc",     500,0,10);
  TH1 *h_sig_5     = new TH1D("h_sig_5",    "5 nuc",     500,0,10);
  TH1 *h_sig_6     = new TH1D("h_sig_6",    "6 nuc",     500,0,10);
  TH1 *h_sig_7     = new TH1D("h_sig_7",    "7 nuc",     500,0,10);
  TH1 *h_sig_8     = new TH1D("h_sig_8",    "8 nuc",     500,0,10);
  TH1 *h_sig_9     = new TH1D("h_sig_9",    "9 nuc",     500,0,10);

  int n_sv_events = 0;
  for (int l = 0;l < n_sv_lines;l++)
    n_sv_events += arr[3*l + 1];

  SVEvent **svs = new SVEvent*[n_sv_events];
  int index = 0;
  for (int l = 0;l < n_sv_lines;l++) {
    int type = arr[3*l];
    int n_ev = arr[3*l + 1];
    int len  = arr[3*l + 2];
    for (int i = 0;i < n_ev;i++) {
      SVEvent *sv_event = NULL;
      while (!sv_event) {
	sv_event = SVEvent::generateEvent(start,end,len,type);
	if (!sv_event) break;
	int s2 = sv_event->getStart();
	int e2 = sv_event->getEnd();
	TString seq = ref_genome.getSequence(s2,e2);
	if (seq.First('N') >= 0 || seq.First('X') >= 0) {
	  delete sv_event;
	  sv_event = NULL;
	  continue;
	}
	bool overlap = false;
	for (int i = 0;i < index;i++) {
	  int s1 = svs[i]->getStart();
	  int e1 = svs[i]->getEnd();
	  if ((s2 <= s1 && e2 >= s1) || (s2 <= e1 && e2 >= e1) ||
	      (s2 >= s1 && e2 <= e1)) {
	    overlap = true;
	    break;
	  }
	}
	if (overlap) {
	  delete sv_event;
	  sv_event = NULL;
	}
      }
      if (sv_event) {
	h_sv_start->Fill(sv_event->getStart());
	h_sv_end->Fill(sv_event->getEnd());
	h_sv_len->Fill(sv_event->getLength());
	svs[index++] = sv_event;
      }
    }
  }

  Genome sv_genome(ref_genome);
  sv_genome.applyVariations(svs,n_sv_events);

  if (sv_genome.getEnd() > end) end = sv_genome.getEnd();
  int max_frg_size = int(exp(logmean + 4*logsd))/1000*1000;
  TH1 *h_frg_start = new TH1D("h_frg_start","Frg start",1000,start,end);
  TH1 *h_frg_end   = new TH1D("h_frg_end  ","Frg end  ",1000,start,end);
  TH1 *h_frg_len   = new TH1D("h_frg_len  ","Frg len  ",
			      1000,0.5,max_frg_size + 0.5);

  start = sv_genome.getStart();
  end   = sv_genome.getEnd();
  len   = sv_genome.getLength();
  cout<<"SV genome is chr"<<sv_genome.getChromosome()<<":"
      <<start<<"-"<<end<<endl;
  cout<<"It is of "<<sv_genome.getLength()<<" nucleotides in length."<<endl;
  cout<<"Chunked in "<<sv_genome.getNChunks()<<" pieces."<<endl<<endl;

  // Shifter to deduct coordinates from sv genome to the reference one
  GenomeShifter shifter(svs,n_sv_events);

  // Open output files
  ofstream rout("reads_out_pure.fa");
  ofstream rout1("reads_out1.fa");
  ofstream rout2;
  if (useSolexa) rout2.open("reads_out2.fa");

  // Calculated maximal coverage
  long long max = 0, n_covered = 0;
  for (int i = 0;i < coverage;i++) {
    max += sv_genome.getLength();
    if (diploid) max += ref_genome.getLength();
  }

  DNAFragment *chimericFrg1 = NULL, *chimericFrg2 = NULL;
  Sequenator sr;
  int n_all = 0, n_out = 0, n_fail = 0;
  Genome *genome = &sv_genome;
  while (n_covered < max) {

    n_fail++;
    if (n_fail > 1000000) { // Check for long (infinite) looping
      cerr<<"Too many errors. Giving up."<<endl;
      break;
    }

    // Generate fragment and assign sequence
    DNAFragment *frg = DNAFragment::generateFragment(genome,logmean,logsd);

    bool fail = true;
    if (use454) { // 454 sequencing technique
      if (chimeric) { // Make chimeric construct
	if (!chimericFrg1) {
	  chimericFrg1 = frg;
	  continue;
	} else chimericFrg2 = frg;

	// Do chimeric fragment
	TString chimeric_seq = constructChimericSequence(chimericFrg1,
							 chimericFrg2);

	delete chimericFrg1;
	delete chimericFrg2;
	chimericFrg1 = chimericFrg2 = NULL;
	frg = new DNAFragment(chimeric_seq);
      }

      if (!frg->make454Read(reads_distr)) // Cannot make a read
	cerr<<"Can not make a 454 read."<<endl;
      else {
	fail = false;

	TString read = frg->get454ReadSeq(),readReal;
	if (read.First('N') < 0 && read.First('X') < 0) { // Output
	  n_out++;
	  int shift = 0;
	  TString header = "";
	  if (genome == &sv_genome)
	    header = frg->make454Header(n_out,&shifter);
	  else
	    header = frg->make454Header(n_out,NULL);
	  rout<<header<<endl<<read<<endl;
	  readReal = sr.sequence454(read);
	  rout1<<header<<endl<<readReal<<endl;
	}
	if (readReal.Length()  > 0) h_read_len->Fill(readReal.Length());
	else if (read.Length() > 0) h_read_len->Fill(read.Length());
      }
    } else if (useSolexa) { // Solexa sequencing technique
      if (!frg->makeSolexaReads(reads_distr))
	cerr<<"Can not make a Solexa read."<<endl;
      else {
	fail = false;

	TString read1 = frg->getSolexaReadSeq1();
	TString read2 = frg->getSolexaReadSeq2();
	TString readReal1,readReal2;
	if (read1.First('N') < 0 && read2.First('X') < 0 &&
	    read1.First('N') < 0 && read2.First('X') < 0) {
	  n_out++;

	  // First read
	  TString header1 = "";
	  if (genome == &sv_genome)
	    header1 = frg->makeSolexaHeader1(n_out,&shifter);
	  else
	    header1 = frg->makeSolexaHeader1(n_out,NULL);
	  rout<<header1<<endl<<read1<<endl;
	  readReal1 = sr.sequenceSolexa(read1);
	  rout1<<header1<<endl<<readReal1<<endl;
	  
	  // Second read
	  TString header2 = "";
	  if (genome == &sv_genome) 
	    header2 = frg->makeSolexaHeader2(n_out,&shifter);
	  else
	    header2 = frg->makeSolexaHeader2(n_out,NULL);
	  rout<<header2<<endl<<read2<<endl;
	  readReal2 = sr.sequenceSolexa(read2);
	  rout2<<header2<<endl<<readReal2<<endl;
	}

	if (readReal1.Length()  > 0) h_read_len->Fill(readReal1.Length());
	else if (read1.Length() > 0) h_read_len->Fill(read1.Length());
	if (readReal2.Length()  > 0) h_read_len->Fill(readReal2.Length());
	else if (read2.Length() > 0) h_read_len->Fill(read2.Length());
      }
    } else  cerr<<"Don't know how to make a read."<<endl;

    if (fail) {
      delete frg;
      continue;
    }

    n_fail = 0;

    n_covered += frg->getLength();
    n_all++;
    h_frg_start->Fill(frg->getStart());
    h_frg_end->Fill(frg->getEnd());
    h_frg_len->Fill(frg->getLength());
    h_end_len->Fill(frg->getEndLength1());
    h_end_len->Fill(frg->getEndLength2());
    delete frg;

    if (diploid)
      if (genome == &sv_genome) genome = &ref_genome;
      else                      genome = &sv_genome;
  }
  rout1.close();
  rout.close();

  delete reads_distr;

  cout<<"Generated "<<n_all<<" fragments for ";
  if (diploid) cout<<"diploid ";
  else         cout<<"haploid ";
  cout<<"genome."<<endl;
  cout<<"Printed reads for "<<n_out<<" fragments."<<endl;
  cout<<"Average fragment length is "<<double(n_covered)/double(n_all)
      <<" nucleotides."<<endl<<endl;

  // Print error statistics
  cout<<"Sequencing error statistics"<<endl;
  int n_err = sr.getMaxPolymerSize();
  int n_seq = 0, n_ins = 0, n_del = 0, n_sub = 0;
  for (int j = 0;j < n_err;j++) {
    int s = sr.getNSequenced(j);
    int i = sr.getNInsertions(j);
    int d = sr.getNDeletions(j);
    int u = sr.getNSubstitutions(j);
    n_seq += s;
    n_ins += i;
    n_del += d;
    n_sub += u;
    if (j <= 10) {
      cout<<j<<": "<<s<<" "<<i<<" "<<d<<" "<<u<<" ";
      if (s != 0) cout<<double(i + d + u)/double(s)<<endl;
      else        cout<<0<<endl;
    }
  }
  cout<<"Total: "<<n_seq<<" "<<n_ins<<" "<<n_del<<" "<<n_sub<<" ";
  if (n_seq != 0) cout<<double(n_ins + n_del + n_sub)/double(n_seq)<<endl;
  else            cout<<0<<endl;

  //TString test = "ATATATC";
//    TString test = "CCCCC";
//    cout<<test<<endl;
//    cout<<sr.sequence454(test)<<endl;
//    cout<<sr.sequence454(test)<<endl;
//    cout<<sr.sequence454(test)<<endl;
//    cout<<sr.sequence454(test)<<endl;
//    cout<<sr.sequence454(test)<<endl;
//    cout<<sr.sequence454(test)<<endl;
//    cout<<sr.sequence454(test)<<endl;
//    cout<<sr.sequence454(test)<<endl;
//    cout<<sr.sequence454(test)<<endl;
  for (int i = 0;i < 500000;i++) {
    h_sig_null->Fill(sr.generate454BGValue());
    h_sig_1->Fill(sr.generate454SignalValue(1));
    h_sig_2->Fill(sr.generate454SignalValue(2));
    h_sig_3->Fill(sr.generate454SignalValue(3));
    h_sig_4->Fill(sr.generate454SignalValue(4));
    h_sig_5->Fill(sr.generate454SignalValue(5));
    h_sig_6->Fill(sr.generate454SignalValue(6));
    h_sig_7->Fill(sr.generate454SignalValue(7));
    h_sig_8->Fill(sr.generate454SignalValue(8));
    h_sig_9->Fill(sr.generate454SignalValue(9));
  }

  ofstream gout("ref_genome.fa");
  gout<<ref_genome;
  gout.close();

  gout.open("sv_genome.fa");
  gout<<sv_genome;
  gout.close();

  ofstream sout("svs_out.txt");
  for (int i = 0;i < n_sv_events;i++) {
    sout<<svs[i]<<endl;
    delete svs[i];
  }
  delete svs;
  sout.close();

  int last = genome_file.Last('/');
  TString givenFileName = genome_file(last + 1,genome_file.Length() - last);
  TFile *file = new TFile(givenFileName + ".root","update");

  h_sv_start->Write(h_sv_start->GetName(),TObject::kOverwrite);
  h_sv_end->Write(h_sv_end->GetName(),TObject::kOverwrite);
  h_sv_len->Write(h_sv_len->GetName(),TObject::kOverwrite);
  h_frg_start->Write(h_frg_start->GetName(),TObject::kOverwrite);
  h_frg_end->Write(h_frg_end->GetName(),TObject::kOverwrite);
  h_frg_len->Write(h_frg_len->GetName(),TObject::kOverwrite);
  h_end_len->Write(h_end_len->GetName(),TObject::kOverwrite);
  h_read_len->Write(h_read_len->GetName(),TObject::kOverwrite);
  h_sig_null->Write(h_sig_null->GetName(),TObject::kOverwrite);
  h_sig_1->Write(h_sig_1->GetName(),TObject::kOverwrite);
  h_sig_2->Write(h_sig_2->GetName(),TObject::kOverwrite);
  h_sig_3->Write(h_sig_3->GetName(),TObject::kOverwrite);
  h_sig_4->Write(h_sig_4->GetName(),TObject::kOverwrite);
  h_sig_5->Write(h_sig_5->GetName(),TObject::kOverwrite);
  h_sig_6->Write(h_sig_6->GetName(),TObject::kOverwrite);
  h_sig_7->Write(h_sig_7->GetName(),TObject::kOverwrite);
  h_sig_8->Write(h_sig_8->GetName(),TObject::kOverwrite);
  h_sig_9->Write(h_sig_9->GetName(),TObject::kOverwrite);

  file->Close();
  delete file;


  return 0;
}


TString constructChimericSequence(DNAFragment *frg1,DNAFragment *frg2)
{
  TString ret = "";
  int option = gRandom->Uniform(0,3.999999);
  if (option == 0) {
    ret = frg1->getSequence() + frg2->getSequence();
  } else if (option == 1) {
    ret = frg1->getSequence() + complement(frg2->getSequence());
  } else if (option == 2) {
    ret = frg2->getSequence() + frg1->getSequence();
  } else if (option == 3) {
    ret = frg2->getSequence() + complement(frg1->getSequence());
  }
  return ret;
}

TString complement(TString seq)
{
  if (!seq) return "";

  TString ret;
  for (int i = seq.Length() - 1;i >= 0;i--) {
    char c = seq[i];
    if (c == 'A')      ret += 'T';
    else if (c == 'T') ret += 'A';
    else if (c == 'C') ret += 'G';
    else if (c == 'G') ret += 'C';
    else               ret +=   c;
  }
  return ret;
}
