#ifndef __GENOME_HH__
#define __GENOME_HH__

#include "SVEvent.hh"

class Genome
{
  // Chromosome
private: int chrom_;
public:  int getChromosome() { return chrom_; }
 
  // Start nucleotide
private: int start_;
public:  int getStart() { return start_; }

  // End nucleotide
private: int end_;
public:  int getEnd() { return end_; }

  // Genome length
public:  int getLength() { return end_ - start_ + 1; }

  // Error message
private: TString error_;
public:  TString getErrorMessage() { return error_; }
public:  bool    isFine() { return error_ == ""; }

  // Chunks of DNA
private:
  static const int MAX_CHUNKS = 250000;
  static const int CHUNK_SIZE = 1000;
  TString chunks_[MAX_CHUNKS];
  SVEvent **chunk_svs_;
  int n_chunks_;
public:
  int getChunkSize() { return CHUNK_SIZE; }
  int getNChunks()   { return n_chunks_; }
  TString &getChunk(int ind)
  {
    TString ret("");
    if (ind >= 0 && ind < n_chunks_ )
      return chunks_[ind];
    return chunks_[n_chunks_];
  }

  // Constructor and destructor
public:
  Genome(TString fileName) : chrom_(-1), start_(-1), end_(-1), error_("")
  {
    ifstream in(fileName.Data());
    if (!in.is_open()) {
      error_ = "ERROR: Cann't open file '" + fileName + "'.";
      return;
    }

    TString f_line;
    in>>f_line;
    if (!parseFirstLine(f_line)) {
      error_ = "ERROR: First line is not in proper format.";
      return;
    }
    if (end_ < start_) {
      error_ = "ERROR: End of DNA is before the beginning.";
      return;
    }

    n_chunks_ = (end_ - start_ + 1)/CHUNK_SIZE + 1;
    TString tmp;
    int len, index = 0;;
    while (!in.eof()) {
      in>>tmp;
      tmp.ToUpper();
      chunks_[index] += tmp;
      len = chunks_[index].Length();
      while (len > CHUNK_SIZE) {
	chunks_[index + 1] = chunks_[index](CHUNK_SIZE,len - CHUNK_SIZE);
	chunks_[index]     = chunks_[index](0,CHUNK_SIZE);
	index++;
	len = chunks_[index].Length();
      }
      if (index > n_chunks_) {
	error_ = "ERROR: Sequence is large than coordinates provided.";
	return;
      }
    }
    in.close();

    if (chunks_[index].Length() > 0) index++;
    if (index != n_chunks_) {
      if (index < n_chunks_) {
	cerr<<"WARNING: Sequence is shorter than coordinates provided."<<endl;
	n_chunks_ = index;
	end_ = start_ + (n_chunks_ - 1)*CHUNK_SIZE +
	  chunks_[n_chunks_-1].Length();
      } else if (index > n_chunks_) {
	error_ = "ERROR: Sequence is larger than coordinates provided.";
	return;
      }
    }
  }

  ~Genome()
  { }

private:
  bool parseFirstLine(TString line)
  {
    if (!line.BeginsWith(">chr")) return false;

    // Parsing chromosome number
    int ind = line.First(':');
    if (ind < 0) ind = line.Length();
    TString num = line(4,ind - 4);
    if (!num.IsDigit()) return false;
    chrom_ = num.Atoi();

    // Setting default values
    start_ = end_ = 1;
    TString rest = line(ind + 1,line.Length() - ind);
    if (rest.Length() == 0) return true;

    // Parsing start and end
    ind = rest.First('-');
    if (ind < 0) return false;
    TString num1 = rest(0,ind);
    TString num2 = rest(ind + 1,line.Length() - ind);
    if (!num1.IsDigit() || !num2.IsDigit()) return false;
    start_ = num1.Atoi();
    end_   = num2.Atoi();

    return true;
  }

public:
  void applyVariations(SVEvent **svs,int n_events)
  {
    // In this function svs are assigned to individual chunks and refered as
    // local svs. Coordinates or local svs start from 0, i.e. same as
    // character indexing in a string.

    chunk_svs_ = new SVEvent*[n_chunks_];
    for (int i = 0;i < n_chunks_;i++) chunk_svs_[i] = NULL;

    for (int i = 0;i < n_events;i++) {
      int start = svs[i]->getStart() - start_;
      int end   = svs[i]->getEnd()   - start_;
      int len   = svs[i]->getLength();
      int s_chunk = start/CHUNK_SIZE;
      int e_chunk = end/CHUNK_SIZE;
      start -= s_chunk*CHUNK_SIZE;
      end   -= e_chunk*CHUNK_SIZE;

      // Inversions spanning over chunk boundaries
      if (svs[i]->isInversion() && s_chunk != e_chunk) {
	TString toInvert;
	for (int c = s_chunk;c < e_chunk;c++) { // Getting seq to invert
	  toInvert += chunks_[c](start,chunks_[c].Length() - start);
	  chunks_[c] = chunks_[c](0,start);
	  start = 0;
	}
	toInvert += chunks_[e_chunk](0,end + 1);
	chunks_[e_chunk] = chunks_[e_chunk](end + 1,CHUNK_SIZE - end - 1);

	TString inverted = invertSeq(toInvert);
	start = 0;
	for (int c = s_chunk;c < e_chunk;c++) { // Inserting back
	  len = CHUNK_SIZE - chunks_[c].Length();
	  chunks_[c] += inverted(start,len);
	  start += len;
	}
	len = inverted.Length() - start;
	chunks_[e_chunk] = inverted(start,len) + chunks_[e_chunk];
	continue;
      }

      if (svs[i]->isInsertion()) { // Insertion
	if (start != end) {
	  cerr<<"Invalid insertion event. Skipping ..."<<endl;
	  continue;
	}
	SVEvent *local_event = SVEvent::makeInsertion(start,
						      svs[i]->getLength());
	local_event->setNext(chunk_svs_[s_chunk]);
	chunk_svs_[s_chunk] = local_event;
      } else if (svs[i]->isDeletion() || svs[i]->isInversion()) {
	int c = s_chunk;
	while (c <= e_chunk) {
	  int s = 0, e = CHUNK_SIZE - 1;
	  if (c == s_chunk) s = start;
	  if (c == e_chunk) e = end;
	  SVEvent *local_event = NULL;
	  if (svs[i]->isDeletion()) local_event = SVEvent::makeDeletion(s,e);
	  else                      local_event = SVEvent::makeInversion(s,e);
	  local_event->setNext(chunk_svs_[c]);
	  chunk_svs_[c] = local_event;
	  c++;
	}
      }
    }

    // Applying to individual chunks
    for (int i = 0;i < n_chunks_;i++)
      if (chunk_svs_[i] != NULL) applyChunkVariations(i);
// 	int n_old = chunks_[i].Length();
// 	int n = applyChunkVariations(i);
// 	if (n > 1) {
// 	  cout<<"Events at "<<i<<endl;
// 	  SVEvent *e = chunk_svs_[i];
// 	  while (e != NULL) {
// 	    cout<<n--<<":  "<<e<<endl;
// 	    e = e->getNext();
// 	  }
// 	  int n_new = chunks_[i].Length();
// 	  cout<<"Old and new for "<<i<<" are "<<n_old<<" "<<n_new<<endl;
// 	}

    deleteChunkSVs();

    // Resizing chunks
    int i_take = 0;
    if (n_chunks_ < MAX_CHUNKS) {
      i_take = MAX_CHUNKS;
      for (int i = n_chunks_ - 1;i >= 0;i--)
	chunks_[--i_take] = chunks_[i];
    }
	
    int i_put = 0;
    TString tmp;
    for (;i_take < MAX_CHUNKS;i_take++) {
      tmp += chunks_[i_take];
      int len = tmp.Length();
      while (len > CHUNK_SIZE && i_put <= i_take) {
	chunks_[i_put++] = tmp(0,CHUNK_SIZE);
	tmp = tmp(CHUNK_SIZE,len - CHUNK_SIZE);
	len = tmp.Length();
      }
    }
    int len = tmp.Length();
    while (len > CHUNK_SIZE) {
      chunks_[i_put++] = tmp(0,CHUNK_SIZE);
      tmp = tmp(CHUNK_SIZE,len - CHUNK_SIZE);
      len = tmp.Length();
    }
    chunks_[i_put++] = tmp;
    n_chunks_ = i_put;
    end_ = start_ + (n_chunks_ - 1)*CHUNK_SIZE +
	  chunks_[n_chunks_-1].Length();
  }

private:
  TString invertSeq(TString toInvert)
  {
    TString ret;
    int len = toInvert.Length();
    for (int i = len - 1;i >= 0;i--) {
      char orig = toInvert(i);
      char inv  = 'N';
      if (orig == 'A')      inv = 'T';
      else if (orig == 'C') inv = 'G';
      else if (orig == 'T') inv = 'A';
      else if (orig == 'G') inv = 'C';
      ret += inv;
    }
    return ret;
  }

  void deleteChunkSVs()
  {
    for (int i = 0;i < n_chunks_;i++) {
      SVEvent *e = chunk_svs_[i];
      while (e) {
	SVEvent *prev = e;
	e = e->getNext();
	delete prev;
      }
      delete e;
    }
    delete[] chunk_svs_;
  }

  void sortSVs(SVEvent **svs,int n)
  {
    for (int i = 0;i < n;i++) {
      int indMove = i;
      for (int j = indMove + 1;j < n;j++)
	if (svs[j]->getStart() < svs[indMove]->getStart())
	  indMove = j;
      SVEvent *tmp = svs[i];
      svs[i] = svs[indMove];
      svs[indMove] = tmp;
    }
  }
  
  int applyChunkVariations(int chunk_index)
  {
    int ret = 0;

    int n_svs = 0;
    for (SVEvent *e = chunk_svs_[chunk_index];e != NULL;e = e->getNext())
      n_svs++;

    SVEvent **sv_arr = new SVEvent*[n_svs];
    n_svs = 0;
    for (SVEvent *e = chunk_svs_[chunk_index];e != NULL;e = e->getNext())
      sv_arr[n_svs++] = e;

    if (n_svs > 1) sortSVs(sv_arr,n_svs);

    TString new_chunk;
    int last_index = 0;
    for (int i = 0;i < n_svs;i++) {
      ret++;
      if (sv_arr[i]->isInsertion()) { // Insertion
	int add_len = sv_arr[i]->getStart() - last_index + 1;
	new_chunk += chunks_[chunk_index](last_index,add_len);
	last_index += add_len;
	TString insert;
	char nucs[] = {'A','C','T','G'};
	for (int j = 0;j < sv_arr[i]->getLength();j++) {
	  int index = int(gRandom->Uniform(0.,3.999999));
	  insert += nucs[index];
	}
	new_chunk += insert;
      } else { // Deletion or inversion
	int add_len = sv_arr[i]->getStart() - last_index;
	new_chunk  += chunks_[chunk_index](last_index,add_len);
	last_index += add_len + sv_arr[i]->getLength();
	if (sv_arr[i]->isInversion()) { // Inversion
	  TString toInvert = chunks_[chunk_index](sv_arr[i]->getStart(),
						  sv_arr[i]->getLength());
	  new_chunk += invertSeq(toInvert);
	}
      }
    }
    new_chunk += chunks_[chunk_index](last_index,CHUNK_SIZE - last_index);

    chunks_[chunk_index] = new_chunk;
    
    return ret;
  }

public:
  TString getSequence(int start,int end)
  {
    start -= start_;
    end   -= start_;
    int s_chunk = start/CHUNK_SIZE;
    int e_chunk = end/CHUNK_SIZE;
    start -= s_chunk*CHUNK_SIZE;
    end   -= s_chunk*CHUNK_SIZE;
    TString ret;
    for (int c = s_chunk;c <= e_chunk;c++)
      if (end >= CHUNK_SIZE) {
	ret += chunks_[c](start,chunks_[c].Length() - start);
	start = 0;
	end -= CHUNK_SIZE;
      } else {
	ret += chunks_[c](start,end - start + 1);
      }
    return ret;
  }
};

ostream& operator << (ostream& os,Genome &g)
{
  static int sym_per_line = 50;
  os<<">chr"<<g.getChromosome()<<":"<<g.getStart()<<"-"<<g.getEnd()<<"\n";
  TString tmp;
  for (int i = 0;i < g.getNChunks();i++) {
    tmp += g.getChunk(i);
    for (int len = tmp.Length();len > sym_per_line;len = tmp.Length()) {
      os<<tmp(0,sym_per_line)<<"\n";
      tmp = tmp(sym_per_line,len - sym_per_line);
    }
  }
  
  return os<<tmp<<"\n";
}

#endif
